from hgraph import TS, GlobalState, set_record_replay_model, IN_MEMORY, get_recorded_value
from hgraph.test import eval_node

from hg_systematic.strategy.recording import recordable, set_recording_prefix, set_record_replay_state, RECORDING_ON, \
    reset_record_replay_state, REPLAYING_ON


@recordable
def my_label_category(ts: TS[int]) -> TS[int]:
    return ts

def test_recording():

    with GlobalState() as gs:
        set_record_replay_model(IN_MEMORY)
        set_recording_prefix("test")
        set_record_replay_state("my_label", "category", RECORDING_ON)
        assert eval_node(my_label_category, [1, 2, 3]) == [1, 2, 3]
        assert len(get_recorded_value("my_label", "test.category")) == 3

        reset_record_replay_state()
        set_record_replay_state("my_label", "category", REPLAYING_ON)
        assert eval_node(my_label_category, [2, 3, 4]) == [1, 2, 3]




from hgraph import graph, TS, TSB, combine, nothing, round_

from hg_systematic.index.configuration import StubIndexConfiguration
from hg_systematic.index.pricing_service import price_index_op, IndexResult
from hg_systematic.index.units import IndexStructure
from hg_systematic.operators import price_in_dollars


__all__ = ["price_stub_index"]


@graph(overloads=price_index_op)
def price_stub_index(config: TS[StubIndexConfiguration]) -> TSB[IndexResult]:
    """Returns the level but not the index structure."""
    return combine[TSB[IndexResult]](
        level=round_(price_in_dollars(config.symbol), config.rounding),  # Technically, a level is not a price in dollars but will do for now.
        index_structure=nothing[TSB[IndexStructure]]()
    )


from contextlib import nullcontext
from dataclasses import dataclass
from typing import Callable

from hgraph import graph, TS, combine, map_, TSB, TSS, feedback, \
    union, TSD, dedup, sample, last_modified_date, convert, dispatch, TSL, Size, nothing
from hgraph import if_true, DebugContext

from hg_systematic.index.configuration import SingleAssetIndexConfiguration
from hg_systematic.index.conversion import roll_schedule_to_tsd
from hg_systematic.index.index_utils import get_monthly_rolling_values, monthly_rolling_index
from hg_systematic.index.pricing_service import price_index_op, IndexResult
from hg_systematic.operators import futures_rolling_contracts, price_in_dollars, MonthlyRollingInfo
from hg_systematic.operators._rolling_rules import spread_rolling_contracts


__all__ = [
    "price_monthly_single_asset_index", "MonthlySingleAssetIndexConfiguration",
    "MonthlySpreadSingleAssetIndexConfiguration", "rolling_contract", "rolling_spread_contract" ,
    "set_single_index_debug_on",
]


DEBUG_ON = False


def set_single_index_debug_on():
    global DEBUG_ON
    DEBUG_ON = True


@dataclass(frozen=True)
class MonthlySingleAssetIndexConfiguration(SingleAssetIndexConfiguration):
    """
    A single asset index that rolls monthly.

    roll_period: tuple[int, int]
        The first day of the roll and the last day of the roll.
        On the first day of the roll the index is re-balanced. The target position is deemed to be
        100% of the next contract. The first day can be specified as a negative offset, this will
        start n publishing days prior to the month rolling into. The second say is the last day of the
        roll and must be positive. On this day, the roll should be completed and the index will hold the
        contract specified for that month in the roll schedule.

        The days represent publishing days of the month, not the calendar day. So 1 (roll period day) may represent
        the 3 day of the calendar month if 1 and 2 were weekends.

        NOTE: A roll period cannot overlap with a prior roll period, so [-10,20] is not allowed as it would
              result in an overlap.

    roll_schedule: tuple[str, ...]
        The roll schedule for this index. This consists of 12 string entries (one for each month), each entry consists
        of a month (letter) and a single digit number representing the year offset for the roll. This will
        be either 0 or 1. For example: ["H0", ..., "X0", "F1"]
        This is used to indicate what contract should be the target for the month the roll period ends in.
        It is possible to specify the same contract, this will effectively be a non-rolling month then.

    roll_rounding: int
        The precision to round the rolling weights to.
    """
    roll_period: tuple[int, int] = None
    roll_schedule: tuple[str, ...] = None
    roll_rounding: int = 8
    trading_halt_calendar: str = None
    contract_fn: Callable[[str, int, int], str] = None


@dataclass(frozen=True)
class MonthlySpreadSingleAssetIndexConfiguration(MonthlySingleAssetIndexConfiguration):
    """
    The spread uses the near leg to be the roll_schedule, the far leg is defined using the far leg roll schedule.
    """
    far_leg_roll_schedule: tuple[str, ...] = None
    contract_fn: Callable[[str, int, int, int, int], str] = None


@dispatch(on=("config",))
def rolling_contract(config: TS[MonthlySingleAssetIndexConfiguration], asset: TS[str],
                     roll_info: TSB[MonthlyRollingInfo]) -> TSL[TS[str], Size[2]]:
    roll_schedule = roll_schedule_to_tsd(config.roll_schedule)

    return futures_rolling_contracts(
        roll_info,
        roll_schedule,
        asset,
        config.contract_fn
    )


@graph(overloads=rolling_contract)
def rolling_spread_contract(config: TS[MonthlySingleAssetIndexConfiguration], asset: TS[str],
                     roll_info: TSB[MonthlyRollingInfo]) -> TSL[TS[str], Size[2]]:
    roll_schedule = roll_schedule_to_tsd(config.roll_schedule)
    far_roll_schedule = roll_schedule_to_tsd(config.far_leg_roll_schedule)

    return spread_rolling_contracts(
        roll_info,
        roll_schedule,
        far_roll_schedule,
        asset,
        config.contract_fn
    )


@graph(overloads=price_index_op)
def price_monthly_single_asset_index(config: TS[MonthlySingleAssetIndexConfiguration]) -> TSB[IndexResult]:
    """
    Support for a monthly rolling single asset index pricing logic.
    For now use the price_in_dollars service to get prices, but there is no reason to use specifically dollars as
    the index just needs a price, it is independent of the currency or scale.
    """
    with nullcontext() if DEBUG_ON is False else DebugContext("[SingleIndex]"):
        DebugContext.print("Starting", config.symbol)
        DebugContext.print("Asset", asset := config.asset)

        # We need the roll_info to compute the contracts, so we get it here
        roll_info, roll_weight = get_monthly_rolling_values(config)

        contracts = rolling_contract(config, asset, roll_info)
        DebugContext.print("contracts", contracts)

        dt = roll_info.dt

        required_prices_fb = feedback(TSS[str], frozenset())
        # Join current positions + roll_in / roll_out contract, perhaps this could be reduced to just roll_in?
        all_contracts = union(combine[TSS[str]](*contracts), required_prices_fb())
        DebugContext.print("all_contracts", all_contracts)

        prices = map_(lambda key, dt_: sample(if_true(dt_ >= last_modified_date(p := price_in_dollars(key))), p),
                      __keys__=all_contracts, dt_=dt)
        DebugContext.print("prices", prices)

        out = monthly_rolling_index(
            config,
            prices,
            compute_target_units_fn=lambda tsb: convert[TSD](target_contract := tsb.contracts[1],
                                                             tsb.level / tsb.prices[target_contract]),
            roll_weight=roll_weight,  # Supplied to reduce unnecessary computation.
            roll_info=roll_info,  # Supplied to reduce unnecessary computation.
            contracts=contracts,
        )

        # We require prices for the items in the current position at least
        required_prices_fb(out.index_structure.current_position.units.key_set)

        return out



"""
Representing the units held (or desired to be held) by an index.
"""
from dataclasses import dataclass

from hgraph import TSD, TS, TimeSeriesSchema, graph, subscription_service, TSS, service_impl, mesh_, TSB

__all__ = ["NotionalUnits", "NotionalUnitValues", "IndexPosition", "IndexStructure",]

# A dictionary of fractional units representing the current or desired holding of the unit
NotionalUnits = TSD[str, TS[float]]

# A dictionary of floating point numbers that represents the value of the units (this can be in notional in the case of
# a SingleAssetIndex or levels in the case of a MultiIndex).
NotionalUnitValues = TSD[str, TS[float]]
# NOTE: Each time an index is modified the value of the units needs to be set correctly.


@dataclass
class IndexPosition(TimeSeriesSchema):
    """
    The units and value of the units as well as the level of the index
    when the units were set.
    """
    units: NotionalUnits
    unit_values: NotionalUnitValues
    level: TS[float]


@dataclass
class IndexStructure(TimeSeriesSchema):
    """
    This three-phase structure allows us to represent the state of the index as well as track the roll period.
    We set the previous_position to the current_position values on the re-balance signal as well as setting the
    target_position to the desired position we wish the index to hold after the roll period.

    Once the roll is completed, these can be released (or left as is if desired).

    We price the index based on the current position and the difference between the current value of the underlyer
    and the current_position's value.

    The value does not change when no re-balancing is occurring.
    """
    current_position: TSB[IndexPosition]
    target_units: NotionalUnits
    previous_units: NotionalUnits





from datetime import date
from typing import Callable, Generic

from hgraph import TSD, TS, Size, operator, SIZE, graph, map_, \
    if_then_else, pass_through, TimeSeriesSchema, TSB, SCALAR, reduce, div_, \
    DivideByZero, passive, union, flip, feedback, ts_schema, sample, component, CompoundScalar, compute_node, STATE

from hg_systematic.operators._calendar import business_day, calendar_for, HolidayCalendar, filter_by_calendar
from hg_systematic.operators._price import price_in_dollars

__all__ = [
    "INDEX_ROLL_FLOAT", "INDEX_ROLL_STR", "index_composition", "index_rolling_contracts",
    "index_rolling_weight", "symbol_is", "index_level"
]

"""
An attempt to describe index computations in a mostly generic way, for now we will constraint
the solution to indices that roll from one underlyer to another (i.e. only two active underlyers for an asset
at any one time)
"""


class _IndexRollingOutput(TimeSeriesSchema, Generic[SCALAR]):
    """
    An output structure representing a floating point values for an operator. The keys are the assets making up the
    index, the values represent the first and second futures involved in the roll. The values are aligned to the
    contracts defined in ``index_rolling_contracts``.
    """
    first: TSD[str, TS[SCALAR]]
    second: TSD[str, TS[SCALAR]]


INDEX_ROLL_FLOAT = TSB[_IndexRollingOutput[float]]
INDEX_ROLL_STR = TSB[_IndexRollingOutput[str]]


def symbol_is(symbol: str, sz: SIZE = Size[2]) -> Callable[[dict, dict], bool]:
    """
    Used to match the ``requires`` rule in the ``overloads`` implementation
    For example:

    ::

        @graph(overloads=index_composition, requires=symbol_is("BCOM Index"))
        def bcom_index_composition(symbol: str, dt: TS[date]) -> INDEX_ROLLING_OUTPUT:
            ...
    """
    return lambda m, s: s["symbol"] == symbol


@operator
def index_composition(symbol: str, dt: TS[date], calendar: HolidayCalendar) -> INDEX_ROLL_FLOAT:
    """
    The index composition for the given date for the given symbol. This composition is used
    as the weighting to compute the level of the assets from the given contracts and their prices.
    """


@operator
def index_rolling_weight(symbol: str, dt: TS[date], calendar: HolidayCalendar) -> TS[float]:
    """
    For a date what is the rolling weight between first and second contracts.
    This is the rolling weight and not the asset weights. With 1.0 representing all
    the weight on the first contract and 0.0 representing all the weight on the second.
    This must align with the ``index_rolling_contracts`` determination of first and second contracts.

    It is useful to build a rolling weight using the ``if_cmp`` operator.
    Consider:

    ::

        is_rolling: TS[CmpResult]
        return if_cmp(is_rolling, 1.0, roll_value, 0.0)

    Alternatively, this can be achieved using a switch:

    ::

        is_rolling: TS[CmpResult]
        return switch_(
            is_rolling,
            {
                CmpResult.LT: lambda ...: 1.0,
                CmpResult.EQ: lambda ...: ...,  # Compute rolling weight
                CmpResult.GT: lambda ...: 0.0,
            },  ...)
    """


@operator
def index_rolling_contracts(symbol: str, dt: TS[date], calendar: HolidayCalendar) -> INDEX_ROLL_STR:
    """
    The first and second contracts for this rolling period.
    For example, for a monthly rolling period, the first and second contracts could be:

    ::
        {
            'ZC': {
                'first': 'ZCZ23',
                'second': 'ZCH24',
            },
            ...
        }

    Then, if there is no roll in the rolling period, it may take the form of:

    ::

        {
            'ZC': {
                'first': 'ZCH24',
                'second': 'ZCH24',
            },
            ...
        }

    This would indicate a non-roll in this rolling period.

    NOTE: The alignment of all rolling operators for a given instrument must be observed to get correct behaviour.
    """


@graph
def weighted_average_value(
        weights: TSD[str, TS[float]],
        contracts: TSD[str, TS[str]],
        prices: TSD[str, TS[float]]
) -> TS[float]:
    """
    Compute the weighted average value for the given weights, contracts and prices.
    """
    values = map_(lambda w, c, p: w * p[c], weights, contracts, pass_through(prices))
    return reduce(lambda x, y: x + y, values, 0.0)


_ComputeIndexLevelsOther = ts_schema(
    new_period=TS[bool],
    rolling_weight=TS[float],
    level_prev=TS[float],
    wav_first_prev=TS[float],
    wav_second_prev=TS[float],
)

_ComputeIndexLevelsReturn = ts_schema(level=TS[float], wav_first=TS[float], wav_second=TS[float])


def compute_index_levels(
        weights_first: TSD[str, TS[float]],
        weights_second: TSD[str, TS[float]],
        contracts_first: TSD[str, TS[str]],
        contracts_second: TSD[str, TS[str]],
        prices: TSD[str, TS[float]],
        other: TSB[_ComputeIndexLevelsOther],
) -> TSB[_ComputeIndexLevelsReturn]:
    # NOTE: We bundle up the single value ticks into a tsb to make recording
    # more simplistic as it will group up the results into a single dataframe.
    new_period = other.new_period
    rolling_weight = other.rolling_weight
    level_prev = other.level_prev
    wav_first_prev = other.wav_first_prev
    wav_second_prev = other.wav_second_prev
    wav_first = weighted_average_value(weights_first, contracts_first, prices)
    wav_second = weighted_average_value(weights_second, contracts_second, prices)
    first_rw = rolling_weight
    second_rw = 1.0 - rolling_weight

    # If we are in the first day of the new period, the first previous is actually
    # The second previous since we now consider the structure from second to
    # now be the structure of first.
    wav_first_prev = if_then_else(new_period, wav_second_prev, wav_first_prev)

    # We mark the previous value contributions as passive to ensure that the feedback
    # of these values does not compute a new value
    value = passive(level_prev) * div_(
        wav_first * first_rw + wav_second * second_rw,
        passive(wav_first_prev * first_rw + wav_second_prev * second_rw),
        DivideByZero.ONE
    )
    return TSB[_ComputeIndexLevelsReturn].from_ts(level=value, wav_first=wav_first, wav_second=wav_second)


@graph
def index_level(symbol: str, initial_level: float = 100.0, record: str = None,
                rounding_fn: Callable[[TS[float]], TS[float]] = None) -> TS[float]:
    dt = business_day(symbol)
    calendar = calendar_for(symbol)

    weights = index_composition(symbol, dt, calendar)
    rolling_weight = index_rolling_weight(symbol, dt, calendar)
    contracts = index_rolling_contracts(symbol, dt, calendar)

    all_contracts = union(flip(contracts.first).key_set, flip(contracts.second).key_set)
    prices = map_(lambda key, d, c: filter_by_calendar(price_in_dollars(key), c), __keys__=all_contracts, d=dt,
                  c=calendar)

    level_fb = feedback(TS[float], initial_level)
    wav_first_fb = feedback(TS[float], 0.0)
    wav_second_fb = feedback(TS[float], 0.0)

    # We have a new period if the weight was 0.0 and is now 1.0
    # We default to the previous state being 0.0, this should not make
    # a difference as we initialise the previous wavs with the same value
    new_period = _new_period(rolling_weight, dt)

    # We don't wrap the function immediately and then delay until we know
    # we want to record or not, if we do we wrap with component and set
    # the recorded id to be the record string value. Allowing for named
    # Recording if we want to support multiple instance of index level
    # computation and record the results independently.
    if record:
        fn = component(compute_index_levels, recordable_id=record)
    else:
        fn = graph(compute_index_levels)

    level_output = fn(
        weights.first, weights.second, contracts.first, contracts.second, prices,
        TSB[_ComputeIndexLevelsOther].from_ts(
            new_period=new_period, rolling_weight=rolling_weight, level_prev=level_fb(),
            wav_first_prev=wav_first_fb(), wav_second_prev=wav_second_fb()
        )
    )

    # We want to ensure we only capture a level value when we are expecting a value.
    # This ensures we don't accidentally compute a level we are not expecting.
    # If we want indicative levels we could return a live copy of this as well, but
    # the previous level should always be the official previous level.
    if rounding_fn:
        level = sample(dt, rounding_fn(level_output.level))
    else:
        level = sample(dt, level_output.level)

    level_fb(level)
    wav_first_fb(level_output.wav_first)
    wav_second_fb(level_output.wav_second)

    return level


class _NewPeriodState(CompoundScalar):
    last_weight: float = 1.0


@compute_node
def _new_period(rolling_weight: TS[float], dt: TS[date], _state: STATE[_NewPeriodState] = None, _output: TS[bool] = None) -> TS[bool]:
    """
    This should tick True when the rolling weight goes from 0.0 to 1.0, otherwise this ticks false when either the
    rolling_weight of the dt ticks and the value of the output is True. This should not duplicate True or False values.
    """
    v = _state.last_weight
    if rolling_weight.modified:
        _state.last_weight = (rw :=rolling_weight.value)
        if v == 0.0 and rw == 1.0:
            return True
    if not _output.valid or _output.value:
        return False



from typing import cast, Mapping

from hgraph import compute_node, cmp_, TS, TSB, CmpResult, service_impl, TSS, TSD, default_path, graph, map_, index_of, \
    switch_, len_, lift, const, cast_, if_then_else, explode, round_

from hg_systematic.operators import MonthlyRollingRange, monthly_rolling_weights, business_day, \
    MonthlyRollingWeightRequest, calendar_for, business_days, Periods
from hg_systematic.operators._calendar import next_month
from hg_systematic.operators._rolling_rules import monthly_rolling_info, MonthlyRollingRequest, MonthlyRollingInfo, \
    rolling_schedules

__all__ = ["monthly_rolling_weights_impl", "monthly_rolling_info_service_impl", "rolling_schedules_service_impl", ]


@compute_node(overloads=cmp_)
def cmp_monthly_rolling_range(lhs: TS[int], rhs: TSB[MonthlyRollingRange], _output: TS[CmpResult] = None) \
        -> TS[CmpResult]:
    """
    Determines if the day index is in the range of the monthly rolling range.
    We only map to GT when day_index == end. When we are not in the range, we otherwise map to LT.
    """
    day_index = lhs.value
    first_day = rhs.first_day.value
    start = rhs.start.value
    end = rhs.end.value

    if day_index == end:
        out = CmpResult.GT
    elif (start < 0 and (day_index > first_day or day_index < end)) or \
            (start >= 0 and (day_index > start and day_index < end)):
        out = CmpResult.EQ
    else:
        out = CmpResult.LT
    if _output.valid and _output.value == out:
        return

    return out


@service_impl(interfaces=monthly_rolling_weights)
def monthly_rolling_weights_impl(
        request: TSS[MonthlyRollingWeightRequest],
        business_day_path: str = "",
        calendar_for_path: str = ""
) -> TSD[MonthlyRollingWeightRequest, TS[float]]:
    """
    Provides an implementation of rolling weights over a monthly rolling range.
    This will only handle requests of the form MonthlyRollingWeightRequest.

    This depends on business_day as well as calendar_for. This assumes that the calendar
    name and the business_day name are the same values.
    """
    return map_(
        _monthly_rolling_weight,
        __keys__=request, __key_arg__="request",
        business_day_path=business_day_path,
        calendar_for_path=calendar_for_path,
    )


@graph
def _monthly_rolling_weight(
        request: TS[MonthlyRollingWeightRequest],
        business_day_path: str,
        calendar_for_path: str,
) -> TS[float]:
    rolling_info = monthly_rolling_info(request)

    start_negative = rolling_info.start < 0
    roll_fraction = 1.0 / switch_(
        start_negative,
        {
            True: lambda s, e: cast(float, abs(s) + e),
            False: lambda s, e: cast(float, e - s)
        },
        rolling_info.start,
        rolling_info.end
    )
    range_ = TSB[MonthlyRollingRange].from_ts(
        first_day=rolling_info.first_day,
        start=rolling_info.start,
        end=rolling_info.end
    )
    is_rolling = rolling_info.roll_state
    weight = switch_(
        is_rolling,
        {
            CmpResult.LT: lambda d, r, f: const(1.0),
            CmpResult.EQ: lambda d, r, f: _weight(d, r, f),
            CmpResult.GT: lambda d, r, f: const(0.0),
        },
        rolling_info.day_index,
        range_,
        roll_fraction,
    )

    return round_(weight, request.round_to)


@graph
def _weight(day_index: TS[int], range_: TSB[MonthlyRollingRange], roll_fraction: TS[float]) -> TS[float]:
    # This is only called when we are in the range,
    # so the logic in the ``if_then_else`` clause will work correctly (else branch).
    offset = cast_(
        float,
        if_then_else(
            day_index >= range_.first_day,
            day_index - range_.first_day,
            day_index - range_.start  # This will only happen when start is negative, so we add the abs(start) (-- => +)
        )
    )
    w = 1.0 - offset * roll_fraction
    return w


@service_impl(interfaces=monthly_rolling_info)
def monthly_rolling_info_service_impl(
        request: TSS[MonthlyRollingRequest],
        business_day_path: str = "",
        calendar_for_path: str = ""
) -> TSD[MonthlyRollingRequest, TSB[MonthlyRollingInfo]]:
    return map_(
        lambda key: monthly_rolling_info_impl(key.start, key.end, key.calendar_name, business_day_path,
                                              calendar_for_path),
        __keys__=request,
    )


@graph
def monthly_rolling_info_impl(
        start: TS[int],
        end: TS[int],
        calendar_name: TS[str],
        business_day_path: str = "",
        calendar_for_path: str = ""
) -> TSB[MonthlyRollingInfo]:
    """
    Computes the rolling info for a given range and calendar name.
    """
    dt = business_day(calendar_name, path=business_day_path if business_day_path else default_path)
    calendar = calendar_for(calendar_name, path=calendar_for_path if calendar_for_path else default_path)
    days_of_month = business_days(Periods.Month, calendar, dt)
    day_index = index_of(days_of_month, dt) + 1
    start_negative = start < 0

    first_day_index = switch_(
        start_negative,
        {
            True: lambda s, dom: len_(dom) + s,
            False: lambda s, dom: s
        },
        start,
        days_of_month
    )

    range_ = TSB[MonthlyRollingRange].from_ts(
        first_day=first_day_index,
        start=start,
        end=end
    )

    is_rolling = cmp_(day_index, range_)

    roll_dt = switch_(
        start < 0,
        {
            True: lambda d, di, end: switch_(
                di > end,
                {
                    True: lambda d_: next_month(d_),
                    False: lambda d_: d_
                },
                d
            ),
            False: lambda d, di, end: d
        },
        dt,
        day_index,
        end,
    )

    y1, m1, _ = explode(roll_dt)
    m2 = (m1 % 12) + 1
    ro = m2 < m1
    y2 = if_then_else(ro, y1 + 1, y1)

    y, m, d = explode(dt)

    begin_roll = first_day_index == day_index
    end_roll = day_index == end

    return TSB[MonthlyRollingInfo].from_ts(
        first_day=first_day_index,
        start=start,
        end=end,
        days_of_month=days_of_month,
        day_index=day_index,
        dt=dt,
        day=d,
        month=m,
        year=y,
        begin_roll=begin_roll,
        end_roll=end_roll,
        roll_state=is_rolling,
        roll_out_month=m1,
        roll_out_year=y1,
        roll_in_month=m2,
        roll_in_year=y2
    )


@service_impl(interfaces=rolling_schedules)
def rolling_schedules_service_impl(
        schedules: Mapping[str, Mapping[int, tuple[int, int]]]
) -> TSD[str, TSD[int, TS[tuple[int, int]]]]:
    """Simple const implementation of rolling_schedules."""
    return const(schedules, TSD[str, TSD[int, TS[tuple[int, int]]]])



from dataclasses import dataclass
from datetime import date
from typing import Mapping

from frozendict import frozendict as fd
from hgraph import CompoundScalar, compute_node, TS, TSB, graph, switch_, dispatch, const, convert, TSD, \
    map_, combine, TSS, reduce, add_, take, nothing, dedup, div_, DivideByZero

from hg_systematic.index.units import IndexStructure

__all__ = ["IndexConfiguration", "StubIndexConfiguration", "BaseIndexConfiguration", "SingleAssetIndexConfiguration",
           "MultiIndexConfiguration", "initial_structure_from_config"]

from hg_systematic.operators import price_in_dollars, trade_date


@dataclass(frozen=True)
class IndexConfiguration(CompoundScalar):
    symbol: str
    rounding: int = 8
    start_date: date = None  # Required to ensure we don't try and compute an index before it has started.
    publish_holiday_calendar: str = None  # Required to know when we can publish a value for an index.


@dataclass(frozen=True)
class BaseIndexConfiguration(IndexConfiguration):
    """

    publish_holiday_calendar: str
        The calendar to use for publishing the index.

    rounding: int
        The number of decimal places to round the published result to

    initial_level: float
        The level to start the index at.

    start_date: date
        The first date of the index. Since the level is path dependent, the start date is required.
    """
    symbol: str
    initial_level: float = 100.0
    current_position: Mapping[str, float] = None
    current_position_value: Mapping[str, float] = None
    current_level: float = 100.0
    target_position: Mapping[str, float] = None
    previous_position: Mapping[str, float] = None


@dataclass(frozen=True)
class StubIndexConfiguration(IndexConfiguration):
    """
    A stub index is one that has a price that can be retrieved using a generic price service.
    This does not provide any initial conditions.
    """


@dataclass(frozen=True)
class SingleAssetIndexConfiguration(BaseIndexConfiguration):
    """
    In order to set appropriate initial conditions, the position data is available to be set.

    asset: str
        The asset symbol. Used to construct the contract name.

    initial_contract: str
        The initial contract to consider as holding on day one. This is used to key the current position and
        the price to construct the percentage holding to compute the initial level.
    """
    asset: str = None
    initial_contract: str = None


@dataclass(frozen=True)
class MultiIndexConfiguration(BaseIndexConfiguration):
    indices: tuple[str, ...] = None


@graph
def initial_structure_from_config(config: TS[IndexConfiguration]) -> TSB[IndexStructure]:
    td = trade_date()
    dt = take(td, 1)
    # This should only compute something on day one. After that it should no longer do anything.
    out = switch_(
        dt == td,
        {
            True: lambda dt_, config_: switch_(
                dt_ == config_.start_date,
                {
                    False: recover_initial_structure_from_config,
                    True: compute_initial_structure,
                },
                config_,
            ),
            False: lambda dt_, config_: nothing(TSB[IndexStructure]),
        },
        dt,
        config,
    )
    # Ensure we have a stable result when we switch to the nothing branch for consumers of the data.
    return dedup(out)


_DEFAULT_VALUE = fd({
    "current_position": {
        "units": {},
        "unit_values": {},
        "level": 100.0
    },
    "previous_units": {},
    "target_units": {},
})


@dispatch(on=("config",))
def recover_initial_structure_from_config(config: TS[IndexConfiguration]) -> TSB[IndexStructure]:
    """
    Recover the index structure from the current state.
    The config may contain the necessary information.
    By default, the empty structure is returned. This will not make the code happy!
    """
    return const(_DEFAULT_VALUE, TSB[IndexStructure])


@dispatch(on=("config",))
def compute_initial_structure(config: TS[IndexConfiguration]) -> TSB[IndexStructure]:
    """
    Create the initial index structure from the current state.
    The config may contain the necessary information.
    By default, the empty structure is returned. This will not make the code happy!
    """
    return const(_DEFAULT_VALUE, TSB[IndexStructure])


@graph(overloads=compute_initial_structure)
def compute_initial_structure_from_single_asset(config: TS[SingleAssetIndexConfiguration]) -> TSB[IndexStructure]:
    """Computes a structure for a single asset index."""
    asset = convert[TSS](config.initial_contract)
    initial_level = config.initial_level
    # TODO: Should provide a version which is just price or price in native currency, ...
    unit_values = map_(lambda key: price_in_dollars(key), __keys__=asset)
    total_value = reduce(add_, unit_values, 0.0)
    units = map_(lambda r: r, div_(initial_level, total_value, DivideByZero.NONE), __keys__=asset)
    # Wait until we have all prices before ticking out the data as a single tick.
    return combine[TSB[IndexStructure]](
        current_position=combine(
            units=units,
            unit_values=unit_values,
            level=config.initial_level
        ),
        previous_units=(c := const(fd(), TSD[str, TS[float]])),
        target_units=c,
    )


@graph(overloads=compute_initial_structure)
def compute_initial_structure_from_multi_asset(config: TS[MultiIndexConfiguration]) -> TSB[IndexStructure]:
    """Computes the initial index structure from the indices"""
    from hg_systematic.index.pricing_service import price_index
    indices = convert[TSS](config.indices)
    initial_level = config.initial_level
    # TODO: Should provide a version which is just price or price in native currency, ...
    unit_values = map_(lambda key: price_index(key).level, __keys__=indices)
    total_value = reduce(add_, unit_values, 0.0)
    # Compute a simple price-weighted contribution to initialise with.
    units = map_(lambda r: r, initial_level / total_value, __keys__ = indices)
    # Wait until we have all prices before ticking out the data as a single tick.
    return combine[TSB[IndexStructure]](
        current_position=combine(
            units=units,
            unit_values=unit_values,
            level=config.initial_level
        ),
        previous_units=(c := const(fd(), TSD[str, TS[float]])),
        target_units=c,
    )


@compute_node(overloads=recover_initial_structure_from_config)
def recover_initial_structure_from_config_for_base_index(config: TS[BaseIndexConfiguration]) -> TSB[IndexStructure]:
    """
    Prepare the initial structure from the index configuration.
    This will tick once only with the values extracted from the index configuration.
    """
    config.make_passive()
    config: BaseIndexConfiguration = config.value
    if config.current_position is None or config.current_position_value is None:
        raise ValueError(
            "When using the BaseIndexConfiguration default for recovering, the current_position and current_position_value may not be None.")
    return {
        "current_position": {
            "units": config.current_position,
            "unit_values": config.current_position_value,
            "level": config.current_level
        },
        "previous_units": {} if config.previous_position is None else config.previous_position,
        "target_units": {} if config.target_position is None else config.target_position,
    }
